==> mars-0.9.2.tar.gz: mars-0.9.2/src/mpu/lib/mars_task_barrier.c <==
/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified
 *  version of the Library in object code or as an executable provided
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete
 *   corresponding machine-readable source code for the modified version
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer
 *   for a complete machine-readable copy of the corresponding source
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#include "mars/mars_task_barrier.h"
#include "mars/mars_task_types.h"
#include "mars/mars_syscalls.h"
#include "mars/mars_mutex.h"
#include "mars/mars_error.h"
#include "mars/mars_debug.h"

static struct mars_task_barrier barrier;

int mars_task_barrier_initialize(uint64_t barrier_ea, uint32_t total)
{
	MARS_CHECK_RET(barrier_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((barrier_ea & MARS_TASK_BARRIER_ALIGN_MASK) == 0,
			MARS_ERROR_ALIGN);
	MARS_CHECK_RET(total <= MARS_TASK_BARRIER_WAIT_MAX, MARS_ERROR_PARAMS);

	mars_mutex_initialize(barrier_ea);
	mars_mutex_lock_get(barrier_ea, (struct mars_mutex *)&barrier);

	barrier.mars_context_ea = mars_get_mars_context_ea();
	barrier.total = total;
	barrier.notified_count = 0;
	barrier.waited_count = 0;
	barrier.notify_wait_count = 0;
	barrier.wait_count = 0;

	mars_mutex_unlock_put(barrier_ea, (struct mars_mutex *)&barrier);

	return MARS_SUCCESS;
}

int mars_task_barrier_notify(uint64_t barrier_ea)
{
	MARS_CHECK_RET(barrier_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((barrier_ea & MARS_TASK_BARRIER_ALIGN_MASK) == 0,
			MARS_ERROR_ALIGN);

	int i;
	struct mars_task_context *task;

	task = (struct mars_task_context *)mars_get_workload_context();

	/* make sure task context has a context save area */
	MARS_CHECK_RET(task->context_save_size && task->context_save_area,
			MARS_ERROR_FORMAT);

	mars_mutex_lock_get(barrier_ea, (struct mars_mutex *)&barrier);

	/* previous barrier wait not complete so wait */
	if (barrier.notified_count == barrier.total) {
		/* add id to wait list */
		barrier.notify_wait_id[barrier.notify_wait_count] =
			mars_get_workload_id();
		barrier.notify_wait_count++;

		mars_mutex_unlock_put(barrier_ea,
				(struct mars_mutex *)&barrier);

		/* wait for signal */
		mars_signal_wait();

		mars_mutex_lock_get(barrier_ea, (struct mars_mutex *)&barrier);
	}

	/* increment notified count */
	barrier.notified_count++;

	/* notified count reached total so release barrier */
	if (barrier.notified_count == barrier.total) {
		/* signal all task ids in wait list */
		for (i = 0; i < barrier.wait_count; i++)
			mars_signal_send(barrier.wait_id[i]);
		barrier.wait_count = 0;
	}

	mars_mutex_unlock_put(barrier_ea, (struct mars_mutex *)&barrier);

	return MARS_SUCCESS;
}

int mars_task_barrier_try_notify(uint64_t barrier_ea)
{
	MARS_CHECK_RET(barrier_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((barrier_ea & MARS_TASK_BARRIER_ALIGN_MASK) == 0,
			MARS_ERROR_ALIGN);

	int i;

	mars_mutex_lock_get(barrier_ea, (struct mars_mutex *)&barrier);

	/* previous barrier wait not complete so return busy */
	if (barrier.notified_count == barrier.total) {
		mars_mutex_unlock_put(barrier_ea,
				(struct mars_mutex *)&barrier);
		return MARS_ERROR_BUSY;
	}

	/* increment notified count */
	barrier.notified_count++;

	/* notified count reached total so release barrier */
	if (barrier.notified_count == barrier.total) {
		/* signal all task ids in wait list */
		for (i = 0; i < barrier.wait_count; i++)
			mars_signal_send(barrier.wait_id[i]);
		barrier.wait_count = 0;
	}

	mars_mutex_unlock_put(barrier_ea, (struct mars_mutex *)&barrier);

	return MARS_SUCCESS;
}

int mars_task_barrier_wait(uint64_t barrier_ea)
{
	MARS_CHECK_RET(barrier_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((barrier_ea & MARS_TASK_BARRIER_ALIGN_MASK) == 0,
			MARS_ERROR_ALIGN);

	int i;
	struct mars_task_context *task;

	task = (struct mars_task_context *)mars_get_workload_context();

	/* make sure task context has a context save area */
	MARS_CHECK_RET(task->context_save_size && task->context_save_area,
			MARS_ERROR_FORMAT);

	mars_mutex_lock_get(barrier_ea, (struct mars_mutex *)&barrier);

	/* check if barrier wait limit reached */
	MARS_CHECK_CLEANUP_RET(barrier.wait_count < barrier.total,
		mars_mutex_unlock_put(barrier_ea,
				(struct mars_mutex *)&barrier),
		MARS_ERROR_LIMIT);

	/* not all tasks notified barrier so need to wait */
	if (barrier.notified_count != barrier.total) {
		/* add id to wait list */
		barrier.wait_id[barrier.wait_count] = mars_get_workload_id();
		barrier.wait_count++;

		mars_mutex_unlock_put(barrier_ea,
				(struct mars_mutex *)&barrier);

		/* wait for signal */
		mars_signal_wait();

		mars_mutex_lock_get(barrier_ea, (struct mars_mutex *)&barrier);
	}

	/* increment waited count */
	barrier.waited_count++;

	/* all tasks have called wait so reset barrier */
	if (barrier.waited_count == barrier.total) {
		barrier.notified_count = 0;
		barrier.waited_count = 0;

		/* signal all task ids in notify wait list */
		for (i = 0; i < barrier.notify_wait_count; i++)
			mars_signal_send(barrier.notify_wait_id[i]);
		barrier.notify_wait_count = 0;
	}

	mars_mutex_unlock_put(barrier_ea, (struct mars_mutex *)&barrier);

	return MARS_SUCCESS;
}

int mars_task_barrier_try_wait(uint64_t barrier_ea)
{
	MARS_CHECK_RET(barrier_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((barrier_ea & MARS_TASK_BARRIER_ALIGN_MASK) == 0,
			MARS_ERROR_ALIGN);

	int i;

	mars_mutex_lock_get(barrier_ea, (struct mars_mutex *)&barrier);

	/* not all tasks notified barrier so return busy */
	if (barrier.notified_count != barrier.total) {
		mars_mutex_unlock_put(barrier_ea,
				(struct mars_mutex *)&barrier);
		return MARS_ERROR_BUSY;
	}

	/* increment waited count */
	barrier.waited_count++;

	/* all tasks have called wait so reset barrier */
	if (barrier.waited_count == barrier.total) {
		barrier.notified_count = 0;
		barrier.waited_count = 0;

		/* signal all task ids in notify wait list */
		for (i = 0; i < barrier.notify_wait_count; i++)
			mars_signal_send(barrier.notify_wait_id[i]);
		barrier.notify_wait_count = 0;
	}

	mars_mutex_unlock_put(barrier_ea, (struct mars_mutex *)&barrier);

	return MARS_SUCCESS;
}

==> mars-08.06.26.tar.gz: mars-08.06.26/src/mpu/lib/mars_task_barrier.c <==
/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be 
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified 
 *  version of the Library in object code or as an executable provided 
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete 
 *   corresponding machine-readable source code for the modified version 
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer 
 *   for a complete machine-readable copy of the corresponding source 
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#include "mars/mars_task_barrier.h"
#include "mars/mars_workload_types.h"
#include "mars/mars_syscalls.h"
#include "mars/mars_mutex.h"
#include "mars/mars_error.h"
#include "mars/mars_debug.h"

static struct mars_task_barrier barrier;

int mars_task_barrier_initialize(uint64_t barrier_ea, uint32_t total)
{
	MARS_CHECK_RET(barrier_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((barrier_ea & MARS_TASK_BARRIER_ALIGN_MASK) == 0,
			MARS_ERROR_ALIGN);
	MARS_CHECK_RET(total <= MARS_TASK_BARRIER_WAIT_MAX, MARS_ERROR_PARAMS);

	mars_mutex_initialize(barrier_ea);
	mars_mutex_lock_get(barrier_ea, (struct mars_mutex *)&barrier);

	barrier.mars_context_ea = mars_get_mars_context_ea();
	barrier.total = total;
	barrier.count = 0;
	barrier.wait_count = 0;

	mars_mutex_unlock_put(barrier_ea, (struct mars_mutex *)&barrier);

	return MARS_SUCCESS;
}

int mars_task_barrier_notify(uint64_t barrier_ea)
{
	MARS_CHECK_RET(barrier_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((barrier_ea & MARS_TASK_BARRIER_ALIGN_MASK) == 0,
			MARS_ERROR_ALIGN);

	int i;

	mars_mutex_lock_get(barrier_ea, (struct mars_mutex *)&barrier);

	/* set the necessary bits */
	barrier.count++;

	/* barrier count reached total so release barrier */
	if (barrier.count == barrier.total) {
		/* signal all task ids in wait list */
		for (i = 0; i < barrier.total; i++)
			mars_signal_send(barrier.wait_id[i]);

		/* flush wait list */
		barrier.wait_count = 0;
	}

	mars_mutex_unlock_put(barrier_ea, (struct mars_mutex *)&barrier);

	return MARS_SUCCESS;
}

int mars_task_barrier_wait(uint64_t barrier_ea)
{
	MARS_CHECK_RET(barrier_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((barrier_ea & MARS_TASK_BARRIER_ALIGN_MASK) == 0,
			MARS_ERROR_ALIGN);

	struct mars_task_context *task;

	/* make sure task context has a context save area */
	task = &mars_get_workload_context()->type_context.task;
	MARS_CHECK_RET(task->context_save_size && task->context_save_area,
			MARS_ERROR_FORMAT);

	mars_mutex_lock_get(barrier_ea, (struct mars_mutex *)&barrier);

	/* check if barrier wait limit reached */
	MARS_CHECK_CLEANUP_RET(barrier.wait_count < MARS_TASK_BARRIER_WAIT_MAX,
		mars_mutex_unlock_put(barrier_ea,
				(struct mars_mutex *)&barrier),
		MARS_ERROR_LIMIT);

	/* not all tasks notified barrier so need to wait */
	if (barrier.count != barrier.total) {
		/* add id to wait list */
		barrier.wait_id[barrier.wait_count] =
			mars_get_workload_id();
		barrier.wait_count++;

		mars_mutex_unlock_put(barrier_ea,
				(struct mars_mutex *)&barrier);

		/* wait for signal */
		mars_signal_wait();

		return MARS_SUCCESS;
	}

	mars_mutex_unlock_put(barrier_ea, (struct mars_mutex *)&barrier);

	return MARS_SUCCESS;
}

int mars_task_barrier_try_wait(uint64_t barrier_ea)
{
	MARS_CHECK_RET(barrier_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((barrier_ea & MARS_TASK_BARRIER_ALIGN_MASK) == 0,
			MARS_ERROR_ALIGN);

	mars_mutex_lock_get(barrier_ea, (struct mars_mutex *)&barrier);

	/* not all tasks notified barrier so return busy */
	if (barrier.count != barrier.total) {
		mars_mutex_unlock_put(barrier_ea,
				(struct mars_mutex *)&barrier);
		return MARS_ERROR_BUSY;
	}

	mars_mutex_unlock_put(barrier_ea, (struct mars_mutex *)&barrier);

	return MARS_SUCCESS;
}
